@@ -325,16 +325,34 @@ static int ci_first_message (void)
     if (old != -1)
       return (old);
 
-    /* If Sort is reverse and not threaded, the latest message is first.
-     * If Sort is threaded, the latest message is first iff exactly one
-     * of Sort and SortAux are reverse.
+    /* If Sort is threaded, the latest message is first iff exactly one
+     * of Sort and the top-level sorting method are reverse.
      */
-    if (((Sort & SORT_REVERSE) && (Sort & SORT_MASK) != SORT_THREADS) ||
-	((Sort & SORT_MASK) == SORT_THREADS &&
-	 ((Sort ^ SortAux) & SORT_REVERSE)))
-      return 0;
+    if ((Sort & SORT_MASK) == SORT_THREADS)
+    {
+      if ((SortThreadGroups & SORT_MASK) == SORT_AUX)
+      {
+        if ((Sort ^ SortAux) & SORT_REVERSE)
+          return 0;
+        else
+          return (Context->vcount ? Context->vcount - 1 : 0);
+      }
+      else
+      {
+        if ((Sort ^ SortThreadGroups) & SORT_REVERSE)
+          return 0;
+        else
+          return (Context->vcount ? Context->vcount - 1 : 0);
+      }
+    }
+    /* If Sort is reverse and not threaded, the latest message is first. */
     else
-      return (Context->vcount ? Context->vcount - 1 : 0);
+    {
+      if (Sort & SORT_REVERSE)
+        return 0;
+      else
+        return (Context->vcount ? Context->vcount - 1 : 0);
+    }
   }
   return 0;
 }
@@ -766,6 +784,25 @@ int mutt_index_menu (void)
 
     if (!in_pager)
     {
+
+#if defined (USE_SLANG_CURSES) || defined (HAVE_RESIZETERM)
+      while (SigWinch)
+      {
+        do
+        {
+          SigWinch = 0;
+          mutt_resize_screen ();
+        }
+        while (SigWinch);
+
+	/*
+	 * force a real complete redraw.  clrtobot() doesn't seem to be able
+	 * to handle every case without this.
+	 */
+	clearok(stdscr,TRUE);
+      }
+#endif
+
       index_menu_redraw (menu);
 
       /* give visual indication that the next command is a tag- command */
@@ -789,21 +826,6 @@ int mutt_index_menu (void)
                           MuttIndexWindow->cols - 1);
       mutt_refresh ();
 
-#if defined (USE_SLANG_CURSES) || defined (HAVE_RESIZETERM)
-      if (SigWinch)
-      {
-	SigWinch = 0;
-	mutt_resize_screen ();
-	menu->top = 0; /* so we scroll the right amount */
-	/*
-	 * force a real complete redraw.  clrtobot() doesn't seem to be able
-	 * to handle every case without this.
-	 */
-	clearok(stdscr,TRUE);
-	continue;
-      }
-#endif
-
       op = km_dokey (MENU_MAIN);
 
       dprint(4, (debugfile, "mutt_index_menu[%d]: Got op %d\n", __LINE__, op));
@@ -1073,8 +1095,9 @@ int mutt_index_menu (void)
                  Prompt when trying to quit Mutt while there are backgrounded
                  compose sessions in process.
               */
-              mutt_yesorno (_("There are $background_edit sessions. Really quit Mutt?"),
-                            MUTT_NO) != MUTT_YES)
+              mutt_query_boolean (OPTBACKGROUNDCONFIRMQUIT,
+                  _("There are $background_edit sessions. Really quit Mutt?"),
+                  MUTT_NO) != MUTT_YES)
           {
             break;
           }
@@ -1337,7 +1360,7 @@ int mutt_index_menu (void)
 
         else if ((op == OP_MAIN_BROWSE_MAILBOXES) ||
                  (op == OP_MAIN_BROWSE_MAILBOXES_READONLY))
-          mutt_buffer_select_file (folderbuf, MUTT_SEL_FOLDER | MUTT_SEL_BUFFY);
+          mutt_buffer_select_file (folderbuf, MUTT_SEL_MAILBOX | MUTT_SEL_BUFFY);
 
 	else
 	{
@@ -1448,6 +1471,7 @@ int mutt_index_menu (void)
         break;
       }
 
+      case OP_GENERIC_SELECT_ENTRY:
       case OP_DISPLAY_MESSAGE:
       case OP_DISPLAY_HEADERS: /* don't weed the headers */
 
@@ -1501,8 +1525,9 @@ int mutt_index_menu (void)
 	{
           if (mutt_background_has_backgrounded () &&
               option (OPTBACKGROUNDCONFIRMQUIT) &&
-              mutt_yesorno (_("There are $background_edit sessions. Really quit Mutt?"),
-                            MUTT_NO) != MUTT_YES)
+              mutt_query_boolean (OPTBACKGROUNDCONFIRMQUIT,
+                  _("There are $background_edit sessions. Really quit Mutt?"),
+                  MUTT_NO) != MUTT_YES)
           {
             break;
           }
@@ -2046,6 +2071,11 @@ int mutt_index_menu (void)
 
 	if (CURHDR->collapsed)
 	{
+          /* Note this returns the *old* virtual index of the root message.
+           *
+           * For sort=reverse-threads this trick allows uncollapsing a
+           * single thread to position on the first (not root) message
+           * in the thread */
 	  menu->current = mutt_uncollapse_thread (Context, CURHDR);
 	  mutt_set_virtual (Context);
 	  if (option (OPTUNCOLLAPSEJUMP))
@@ -2053,8 +2083,22 @@ int mutt_index_menu (void)
 	}
 	else if (option (OPTCOLLAPSEUNREAD) || !UNREAD (CURHDR))
 	{
-	  menu->current = mutt_collapse_thread (Context, CURHDR);
+	  HEADER *base;
+	  int final;
+          /* This also returns the *old* virtual index of the root, but now
+           * we have to find the new position of the root, which isn't
+           * the same for sort=reverse-threads. */
+          final = mutt_collapse_thread (Context, CURHDR);
+	  base = Context->hdrs[Context->v2r[final]];
 	  mutt_set_virtual (Context);
+	  for (j = 0; j < Context->vcount; j++)
+	  {
+	    if (Context->hdrs[Context->v2r[j]]->index == base->index)
+	    {
+	      menu->current = j;
+	      break;
+	    }
+	  }
 	}
 	else
 	{
@@ -2637,9 +2681,10 @@ int mutt_index_menu (void)
         break;
 #endif
 
-      default:
+      case OP_NULL:
 	if (!in_pager)
 	  km_error_key (MENU_MAIN);
+        break;
     }
 
     if (in_pager)
